import operator
from itertools import chain
from collections import defaultdict
import numpy


import wrapper
from ..constants import *
from ..utils.multi_visitor import VisitorFactory, F_CACHE, NF_ERROR, NF_ELSE
from ..itypes import rtypes
from ..passes import required
from .. import engines

_delay_import_(globals(),"..representor")
_delay_import_(globals(),"..utils","util","nestutils","cutils")
_delay_import_(globals(),"..slices")
_delay_import_(globals(),"..itypes","detector","type_attribute_freeze","convertors")
_delay_import_(globals(),"..repops_slice")
_delay_import_(globals(),"..repops_dim")


class PyRepresentor(wrapper.SourceRepresentor):
    pass

def rep(data=None, dtype=None, unpack=True, name=None):#{{{
    if(not dtype is None):
        if(isinstance(dtype,str)):
            dtype = rtypes.createType(dtype)
    else:
        det = detector.Detector()
        det.process(data)
        dtype = det.getType()
    
    if(name is None):
        name = "data"

    data_slice = slices.DataSlice(data,name=name,rtype=dtype)
    data_slice = slices.ensure_normal_or_frozen(data_slice)
    
    while(unpack and data_slice.type.__class__ is rtypes.TypeArray):
        data_slice = slices.UnpackArraySlice(data_slice)

    res = wrapper.SourceRepresentor((data_slice,)) 

    if(unpack and isinstance(res.getType(), rtypes.TypeTuple)):
        res = repops_slice.unpack_tuple(res)
    
    return res#}}}



def var(name, example_data=None, dtype=rtypes.unknown, 
        type_check=False, unpack=True):#{{{
    if(isinstance(dtype, str)):
        dtype = rtypes.createType(dtype)
    if(isinstance(example_data, representor.Representor)):
        assert dtype == rtypes.unknown, \
            "Example representor and dtype cannot both be supplied"
        res = example_data.copy()
    else:
        res = rep(example_data, dtype, type_check=type_check, unpack=unpack)
    res.__class__ = VarPyRepresentor
    res._name = name
    res._type_check = type_check
    del res._data
    return res#}}}


class PySelectIndicator(wrapper.Indicator):
    pass

class Result(object):
    __slots__ = ['data', 'ntypes', 'ndims']
    def __init__(self, data=None, ntypes=None, ndims=None):
        if(ntypes is None):
            self.ntypes = {}
        else:
            assert isinstance(ntypes, dict), "Ntypes attribute should be a dict"
            self.ntypes = ntypes

        if(ndims is None):
            self.ndims = {}
        else:
            assert isinstance(ndims, dict), "Ndims attribute should be a dict"
            self.ndims = ndims

        if(data is None):
            data = {}
        else:
            assert isinstance(data, dict), "Data attribute should be dict"
            self.data = data

    def copy(self):
        return Result(self.data.copy(), self.ntypes.copy(), self.ndims.copy())

    def __repr__(self):
        return "\nRESULT\nData: " + str(self.data) + \
                       "\nNtypes:   " + str(self.ntypes) + \
                       "\nNdims:   " + str(self.ndims) + "\n"

def merge_results(*results):
    ndata = {}
    ntypes = {}
    ndims = {}
    for result in results:
        ndata.update(result.data)
        ntypes.update(result.ntypes)
        ndims.update(result.ndims)
    return Result(ndata, ntypes, ndims)

class PySelectExecutor(VisitorFactory(prefixes=("rep", ), flags=F_CACHE),
                    VisitorFactory(prefixes=("sliceop", "op"), flags=NF_ELSE),
                    wrapper.Executor):
    after = set([required.RequiredSliceIds])

    @classmethod
    def run(cls, query, pass_results):
        self = cls()
        self.query = query
        #pylint: disable-msg=E1101
        return (self.rep(query.root),)
    
    def getSliceById(self, sid, crep, res):
        slice = crep._all_slices[sid]
        if(not sid in res.data):
            if(slice.last_id in res.data):
                res.data[sid] = res.data[slice.last_id]
                if(slice.last_id in res.ntypes):
                    res.ntypes[sid] = res.ntypes[slice.last_id]
                if(slice.last_id in res.ndims):
                    res.ndims[sid] = res.ndims[slice.last_id]
            else:
                #pylint: disable-msg=E1101
                (ndata, ntype, ndims) = self.sliceop(slice, crep, res)
                res.data[slice.id] = ndata
                if(slice.type.__class__ is rtypes.TypeUnknown and
                    not ntype.__class__ is rtypes.TypeUnknown):
                    res.ntypes[slice.id] = ntype
                    res.ndims[slice.id] = ndims
 
        ftype = res.ntypes.get(sid, slice.type)
        fdims = res.ndims.get(sid, slice.dims)
        return (res.data[sid], ftype, fdims)

    def repRepresentor(self, crep):
        #pylint: disable-msg=E1101
        result = self.op(crep).copy()
        
        for rid in crep._req_ids:
            #inlining of getSliceByID
            if(rid in result.data):
                continue
            slice = crep._all_slices[rid]
            if(slice.last_id in result.data):
                result.data[rid] = result.data[slice.last_id]
                if(slice.last_id in result.ntypes):
                    result.ntypes[rid] = result.ntypes[slice.last_id]
                if(slice.last_id in result.ndims):
                    result.ndims[rid] = result.ndims[slice.last_id]
                continue
            #pylint: disable-msg=E1101
            (ndata, ntype, ndims) = self.sliceop(slice, crep, result)
            result.data[slice.id] = ndata
            if(slice.type.__class__ is rtypes.TypeUnknown and
                not ntype.__class__ is rtypes.TypeUnknown):
                result.ntypes[slice.id] = ntype
                result.ndims[slice.id] = ndims
        
        #some memory saving (not necessary)
        for did in result.data.keys():
            if(not did in crep._req_ids):
                del result.data[did]
        return result 
    
    def repfixate(self, crep):
        #pylint: disable-msg=E1101
        res = self.op(crep)
        xres = {}
        for slice in crep._active_slices:
            xres[slice.id] = res.data[slice.id]
        return xres

    def opPyRepresentor(self, crep):
        return Result(crep._data.copy())
    
    def opVarPyRepresentor(self, crep):
        assert crep._name in self.query.args, \
            "Variable " + str(crep._name) + " not found in arguments."
        data = self.query.args[crep._name]
        if(isinstance(data, representor.Representor)):
            if(not crep._no_type_check):
                assert data.type <= crep.type, "Argument has incompatible type"
        else:
            data = rep(data, crep.type, no_type_check=crep._no_type_check)

        res = {}
        for vslice, aslice in zip(crep._active_slices, data._active_slices):
            res[vslice.id] = data._data[aslice.id]

        return Result(res)
    
    def opUnaryOpRep(self, crep):
        #pylint: disable-msg=E1101
        return self.rep(crep._sources[0])

    def opMultiOpRep(self, crep):
        return merge_results(*[self.rep(s) for s in crep._sources])
    
    def opRFilter(self, crep):
        result = self.rep(crep._sources[0]).copy()
        for source, cslice, fslices, fdepth, fbcast in \
                             zip(crep._sources[1:], crep._constraint_slices, 
                                   crep._filter_slices, crep._filter_depths, crep._filter_broadcast):
            constraint_res = self.rep(source)
            cdata = constraint_res.data[cslice.id]

            for depth, slice in zip(fdepth, fslices):
                if(slice.id in crep._req_ids):
                    rdata = result.data[slice.id]
                    for d in depth:
                        if(d):
                            nfbcast = ((1,) * d) + fbcast
                        else:
                            nfbcast = fbcast
                        rdata = nestutils.nest_broadcast(nfbcast, [rdata, cdata], [slice.dims, cslice.dims], 
                                                        inner_filter, (slice.type,), slice.dims, vectorize=1 )
                    result.data[slice.id] = rdata
        return result

    def opGroup(self, crep):
        result = self.rep(crep._sources[0]).copy()
        gresult = self.rep(crep._sources[1]) 
       

        gdata = [gresult.data[slice.id] for slice in crep._group_slices]
        gdims = [slice.dims for slice in crep._group_slices]
        bcast = [1] * (len(gdims[0]) - 1)
        groupdims = crep._keep_slices["__group__"][2]
        gidx = nestutils.nest_broadcast(bcast, gdata, gdims, inner_group, (rtypes.TypeInt64(),), groupdims)
        
        nresult = {}
        for id in crep._req_ids:
            nresult[id] = result.data[id]

        args = {'count_sel':None}
        for key, value in crep._keep_slices.iteritems():
            if(key == "__group__"):
                nbcast = (3,) * (len(gdims[0]) - 1) + (2,) * len(gdata)
                gindex = gidx
            else:
                nbcast = (3,) * (len(gdims[0]) - 1) + (2,) * len(key)
                args['key'] = key
                gindex = nestutils.nest_broadcast((1,) * (len(gdims[0]) - 1), (gidx,), (groupdims,), inner_groupcondense, (rtypes.TypeInt64(),), groupdims[:(len(gdims[0]) - 1 + len(key))], func_attr=args)
                
            xslices = value[0]
            depths = value[1]
            xdims = value[2]

            for depth, slice in zip(depths, xslices):
                if(slice.id in crep._req_ids):
                    rdata = nresult[slice.id]
                    for d in depth:
                        if(d):
                            nfbcast = ((1,) * d) + nbcast
                        else:
                            nfbcast = nbcast
                        rdata = nestutils.nest_broadcast(nfbcast, [rdata, gindex], [slice.dims, xdims], 
                                                        inner_filter, (slice.type,), slice.dims, vectorize=1)
                    nresult[slice.id] = rdata

        

        return Result(nresult)
   
    def opflat(self, crep):
        result = self.rep(crep._sources[0]).copy()
       
        sdepths = []
        sdatas = []
        stypes = []
        sids = []
        for sliceid in crep.inner_req_ids:
            slice = crep._all_slices[sliceid]
            sdim = slice.dims
            depth = 0
            for sd, fd in zip(sdim, crep._flat_dim):
                if(sd == fd):
                    depth += 1
                else:
                    break
            if(sliceid in crep._flat_slices):
                depth += 1
            if(depth >= len(crep._flat_dim) - 1):
                sdepths.append(depth)
                sdatas.append(result.data[sliceid])
                if(len(sdim) >= len(crep._flat_dim)):
                    stypes.append(object)
                else:
                    stypes.append(slice.type.toNumpy())
                sids.append(sliceid)
        nsdatas = nestutils.nest_flatten(sdatas, sdepths, stypes)
        for nsdata, sliceid in zip(nsdatas, sids):
            result.data[sliceid] = nsdata
        
        return result

    def opJoin(self, crep):
        s1 = self.rep(crep._sources[0]).copy()
        s2 = self.rep(crep._sources[1]).copy()
        (lsliceids, rsliceids) = crep._join_slices

        
        #determine fields to be joined
        lids = lsliceids & crep._req_ids
        rids = rsliceids & crep._req_ids

        #check if there is at least one field, otherwise add it
        if(not lids):
            lids = [(lsliceids.intersection(s1.data.keys())).pop()]
        if(not rids):
            rids = [(rsliceids.intersection(s2.data.keys())).pop()]
        
        #ensure constant order
        lids = list(lids)
        rids = list(rids)

        #get data to be joined
        ldata = [s1.data[sliceid] for sliceid in lids]
        rdata = [s2.data[sliceid] for sliceid in rids]
         
        res = nestutils.nest_crossjoin(crep._join_broadcast, ldata, rdata, crep._join_group)        
        result = merge_results(s1, s2)
        
        for id, col in zip(chain(lids, rids), res):
            result.data[id] = col
        
        return result

    def opEquiJoin(self, crep):
        s1 = self.rep(crep._sources[0]).copy()
        s2 = self.rep(crep._sources[1]).copy()
        (lsliceids, rsliceids) = crep._join_slices
        
        #determine fields to be joined
        lids = lsliceids & crep._req_ids
        rids = rsliceids & crep._req_ids
        
        #check if there is at least one field, otherwise add it
        if(not lids):
            lids = [(lsliceids.intersection(s1.data.keys())).pop()]
        if(not rids):
            rids = [(rsliceids.intersection(s2.data.keys())).pop()]

        #ensure constant order
        lids = list(lids)
        rids = list(rids)

        #get data to be joined
        ldata = [s1.data[sliceid] for sliceid in lids]
        rdata = [s2.data[sliceid] for sliceid in rids]

        cldata = s1.data[crep._lsliceid]
        crdata = s2.data[crep._rsliceid]
        cldepth = len(crep._sources[0]._all_slices[crep._lsliceid].dims)
        crdepth = len(crep._sources[1]._all_slices[crep._rsliceid].dims)

        res = nestutils.nest_equijoin(crep._join_broadcast, ldata, rdata, cldata, crdata, cldepth, crdepth, crep._join_group)        
        result = merge_results(s1, s2)
        
        for id, col in zip(chain(lids, rids), res):
            result.data[id] = col
        
        return result
       

    def opelse(self, crep):
        assert hasattr(crep, "pyexec"), \
                "Cannot find execution method for " + str(crep.__class__)
        return crep.pyexec(self)
   
    def sliceopSlice(self, slice, crep, res):
        raise RuntimeError, "Cannot find source for slice " + \
                                str(slice) + " (id:" + str(slice.id) + ")"

    def sliceopUnpackArraySlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        assert isinstance(stype, rtypes.TypeArray), \
            "Unpacking array failed: source is not an array but " + str(stype)

        subtype = stype.subtypes[0]
        dconvertor = convertors.getConvertor(subtype)
        ndims = sdims + stype.dims
        func = lambda data, bcast, dims, pack, rtypes, rdims, args: dconvertor.convert(rtypes[0], data[0]) 
        ndata = nestutils.nest_broadcast([1] * len(sdims), [sdata], [sdims], func, [subtype], ndims)
        return (ndata, subtype, ndims)
    
        
    def sliceopUnpackTupleSlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        assert isinstance(stype, rtypes.TypeTuple), \
            "Unpacking tuple failed: source is not an tuple but " + str(stype)
        
        subtype = stype.subtypes[cslice.slice_idx]
        scanner = scanner_protocol.getScanner(subtype)

        inner_func = operator.itemgetter(cslice.slice_idx)
        if(stype.has_missing):
            inner_func = util.filter_missing(inner_func)
        ndata = nestutils.nest_broadcast([1] * (len(sdims) - 1), [sdata], [sdims], inner_detuple, 
                                    [subtype], sdims, func_attr={'selector':inner_func, 'scanner':scanner})
        return (ndata, subtype, sdims)
        

    def sliceopBinElemOpSlice(self, cslice, crep, res):
        data1, type1, dims1 = self.getSliceById(cslice.source_ids[0], crep, res)
        data2, type2, dims2 = self.getSliceById(cslice.source_ids[1], crep, res)
        xfunc = cslice.oper.exec_funcs["py"]
        func = lambda x, y: xfunc(x, y, type1, type2, cslice.type, cslice.op)
        rdata = nestutils.nestop([data1, data2], func, [len(dims1), len(dims2)])
        return (rdata, cslice.type, cslice.dims)
    
    def sliceopUnaryElemOpSlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        func = cslice.oper.exec_funcs["py"]
        xfunc = lambda data, bcast, dim, pack, rtypes, rdims, args: func(data[0], stype, rtypes[0], cslice.op) 
        ndata = nestutils.nest_broadcast([1] * (len(sdims) -1), [sdata], [sdims], xfunc, [cslice.type], cslice.dims, vectorize=1)
        return (ndata, cslice.type, cslice.dims)
       
    def sliceopMapSeqSlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        if(cslice.params or cslice.kwds):
            efunc = cslice.exec_func
            params = cslice.params
            kwds = cslice.kwds
            xfunc = lambda x : efunc(x, *params, **kwds)
        else:
            xfunc = cslice.exec_func
        rdata = nestutils.nestmap(sdata, xfunc, len(sdims) - 1)
        return (rdata, cslice.type, cslice.dims)
    
    def sliceopMapSlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        xfunc = cslice.exec_func
        rdata = nestutils.nestmap(sdata, xfunc, len(sdims), cslice.type.toNumpy())
        return (rdata, cslice.type, cslice.dims)
  
    def sliceopPackTupleSlice(self, cslice, crep, res):
        data = []
        for sourceid in cslice.source_ids:
            sdata, stype, sdims = self.getSliceById(sourceid, crep, res)
            if(cslice.to_python):
                sdata = sdata.tolist()
            data.append(sdata)
        rfunc = lambda *x: tuple(x)
        rdata = nestutils.nestop(data, rfunc, [len(cslice.dims) + 1] * len(data)) 
        return (rdata, cslice.type, cslice.dims)


    def sliceopPackArraySlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        return (sdata, cslice.type, cslice.dims)
    
    def sliceopPackListSlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        sdata = nestutils.nestmap(sdata, lambda x: x.tolist(), len(cslice.dims))
        return (sdata, cslice.type, cslice.dims)

    def sliceopFreezeSlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)

        scanner = scanner_protocol.getScanner(stype)
        if(sdims):
            rdata = data_freezer.freeze(scanner, sdata, stype)
        else:
            rdata = data_freezer.freeze(scanner, [sdata], stype)[0]
      
        return (rdata, cslice.type, cslice.dims)
      
    def sliceopAggregateSlice(self, cslice, crep, res):
        sdata, stype, sdims = self.getSliceById(cslice.source_ids[0], crep, res)
        xfunc = cslice.exec_func
        rdata = nestutils.nestmap(sdata, xfunc, len(sdims) - 1,
                                        cslice.type.toNumpy())
        return (rdata, cslice.type, cslice.dims[:-1])
   

#helper funcs
def inner_detuple(data, bcast, dims, pack, rtypes, rdims, args):
    selector = args['selector']
    if(not rdims):
        return data_convertor.convert(args['scanner'], [selector(data[0])], rtypes[0])[0]
    else:
        return data_convertor.convert(args['scanner'], [selector(elem) for elem in data[0]], rtypes[0]) 


def inner_group(data, bcast, dims, pack, rtypes, rdims, args):
    indexes = []
    if(len(data) > 1):
        for dcol in data:
            index = {}
            pos = 0 
            for elem in dcol:
                if not elem in index:
                    index[elem] = pos
                    pos += 1
            indexes.append(index)
        
        data_dict = defaultdict(list)
        for pos, elems in enumerate(zip(*data)):
            data_dict[elems].append(pos)
       
        shape = [len(index) for index in indexes]
        indexdata = [cutils.darray([],int)] * numpy.prod(shape)
        indexdata = cutils.darray(indexdata, object, 1).reshape(tuple(shape))
        
        for key, posses in data_dict.iteritems():
            loc = tuple([index[keypart] for index, keypart in zip(indexes,key)])
            indexdata[loc] = cutils.darray(posses, int)
    else:
        data_dict = defaultdict(list)
        for pos, elems in enumerate(*data):
            data_dict[elems].append(pos)

        indexdata = [cutils.darray(elem, int) for elem in data_dict.values()]
        indexdata = cutils.darray(indexdata, object, 1)
   
    return indexdata


def inner_groupcondense(data, bcast, dims, pack, rtypes, rdims, args):
    ndim = len(dims[0]) -1
    key = args['key']
    firstitem =operator.itemgetter(0)
    if(ndim > 1):
        if(args['count_sel'] is None):
            args['count_sel'] = cutils.darray([len(elem) for elem in data[0].ravel()],int).reshape(data[0].shape)
        tidx = set(range(ndim))
        tidx = tidx - set(key)
        tidx = list(tidx)
        tidx.sort()
        csel = numpy.transpose(args['count_sel'], key + tuple(tidx))
        rshape = tuple([args['count_sel'].shape[ti] for ti in key]) +  (numpy.prod([args['count_sel'].shape[ti] for ti in tidx]),)
        csel = csel.reshape(rshape)
        idx = numpy.argmax(csel, axis=len(key))
        cgidx = numpy.transpose(data[0], key + tuple(tidx))
        cgidx = cgidx.reshape(rshape)

        other_idx = numpy.indices(rshape[:len(key)])                    
        cgidx = cgidx[tuple(other_idx) + (idx,)]
        gindex = cutils.darray([firstitem(elem) for elem in cgidx.ravel()],int).reshape(cgidx.shape)
    else:
        assert len(key) == 1 and key[0] == 0, "Invalid keep index"
        gindex = cutils.darray([elem[0] for elem in data[0]],int)
    return gindex
       
def inner_filter(data, bcast, dims, pack, rtypes, rdims, args):
    filter_data, constraint_data = data
    if(isinstance(constraint_data, numpy.ndarray) and constraint_data.dtype == object):
        constraint_data = numpy.cast[bool](constraint_data)
    if(len(bcast) > 0):
        constraint_data = (slice(None),) * len(bcast) + (constraint_data,)
    return filter_data[constraint_data]



import urllib2
import urlparse
import os
from os import path
from distutils.dir_util import mkpath

class DownloadCache(object):
    def __init__(self, folder=None):
        if(folder is None):
            folder = path.expanduser("~") + "/.ibidas/data"
        self.set_download_folder(folder)

    def url_to_file(self, url):
        return url.split('/')[-1]

    def __call__(self, url, dbname=None, filename=None):
        if(filename is None):
            filename = self.url_to_file(url)
        if(dbname is None):
            dbname = urlparse.urlparse(url).netloc

        file_path = self.get_filename(dbname, filename) 
        if not path.exists(file_path):
            self.download(url, file_path)
        
        return file_path

    def set_download_folder(self, folder):
        self.folder = folder

    def check_dir(self, dir):
        if not path.isdir(dir):
            assert not path.exists(dir), "Folder: " + str(dir) + " already exists as non-directory"
            mkpath(dir)

    def get_filename(self, dbname, filename):
        self.check_dir(self.folder)
        if(dbname):
            self.check_dir(self.folder + "/" + dbname)
            return self.folder + "/" + dbname + "/" + filename
        else:
            return self.folder + "/" + filename

    def download(self, url, file_name):
        u = urllib2.urlopen(url)
        f = open(file_name, 'w')
        meta = u.info()
        file_size = int(meta.getheaders("Content-Length")[0])
        print "Downloading: %s Bytes: %s" % (file_name, file_size)

        file_size_dl = 0
        block_sz = 8192
        while True:
            buffer = u.read(block_sz)
            if not buffer:
                break

            file_size_dl += block_sz
            f.write(buffer)
            status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
            status = status + chr(8)*(len(status)+1)
            print status,

        f.close()




import operator

from constants import *
import repops

_delay_import_(globals(),"utils","util","context")
_delay_import_(globals(),"ops")
_delay_import_(globals(),"representor")
_delay_import_(globals(),"wrappers","wrapper_py")
_delay_import_(globals(),"itypes","rtypes","dimensions","dimpaths")
_delay_import_(globals(),"repops_funcs")
_delay_import_(globals(),"repops_dim")

class Broadcast(repops.MultiOpRep):
    def _process(self,sources, mode="dim"):
        state = reduce(operator.__and__,[source._state for source in sources])
        if not state & RS_SLICES_KNOWN:
            return

        nslices = []
        for bcslices in util.zip_broadcast(*[source._slices for source in sources]):
            nslices.extend(ops.broadcast(bcslices,mode)[0])
        return self._initialize(tuple(nslices),state)


class Nest(repops.MultiOpRep):
    def __init__(self, lsource, rsource, dim=None):
        repops.MultiOpRep.__init__(self,(lsource,rsource),dim=dim)
    def _process(self, sources, dim=None):
        assert len(sources) == 2, "Nest expects two representor objects"
        state = reduce(operator.__and__,[source._state for source in sources])
        if not state & RS_SLICES_KNOWN:
            return
        lsource,rsource = sources
        
        joinpath = dimpaths.identifyUniqueDimPathSource(lsource, dim)
        joinpath = dimpaths.extendParentDim(joinpath,[s.dims for s in source._slices], ALL)

        idims = []
        for i in xrange(len(joinpath)):
            idims.append(dimensions.Dim(1))
        
        
        references = []
        for ndim in joinpath:
            nrefs = []
            for slice in lsource._slices:
                if ndim in slice.dims:
                    nrefs.append(slice)
            references.append(nrefs)

        nslices = []
        plan = [BCEXIST] * len(idims)
        for slice in rsource._slices:
            odims = slice.dims
            for dimpos in xrange(len(joinpath)):
                slice = ops.InsertDimOp(slice,dimpos,idims[dimpos])
            slice = ops.BroadcastOp(slice,references,plan,joinpath + odims)
            nslices.append(slice)
        return self._initialize(tuple(nslices), state)

class Combine(repops.MultiOpRep):
    def __init__(self, *sources):
        repops.MultiOpRep.__init__(self, sources)
    def _process(self,sources):
        state = reduce(operator.__and__,[source._state for source in sources])
        if not state & RS_SLICES_KNOWN:
            return
        nslices = self._apply(*[source._slices for source in sources])
        return self._initialize(nslices,state)

    @classmethod
    def _apply(cls, *xslicelists):
        if(len(xslicelists) == 2):
            lslices,rslices = xslicelists
            lslices = [ops.ChangeBookmarkOp(lslice,add_bookmark="!L",update_auto_bookmarks="L") for lslice in lslices]
            rslices = [ops.ChangeBookmarkOp(rslice,add_bookmark="!R",update_auto_bookmarks="R") for rslice in rslices]
            return tuple(lslices + rslices)
        else:
            return sum([tuple(xslices) for xslices in xslicelists],tuple())

class Group(repops.MultiOpRep):
    def __init__(self,source,constraint,flat={}):
        repops.MultiOpRep.__init__(self,(source,constraint),flat=flat)

    def _process(self,sources, flat):
        source, gsource = sources 
        if not source._state & RS_SLICES_KNOWN or not gsource._state & RS_TYPES_KNOWN:
            return

        gslices = [ops.ensure_frozen(slice) for slice in gsource._slices]
        gslices = ops.broadcast(gslices,mode="dim")[0]
        gslices = [ops.PackArrayOp(gslice,1) for gslice in gslices]

        gslice = ops.GroupIndexOp(gslices)
        gslice = ops.UnpackArrayOp(gslice, len(gslices))

        #determine which slices to flatten
        oslices = list(source._slices)
        nslices = [None] * len(oslices)
        xflat = {}
        #specified slices
        for key,values in flat.iteritems():
            if(isinstance(key,tuple)):
                keyslices = gsource.Get(*key)
            else:
                keyslices = gsource.Get(key)
            nkey = tuple([gsource._slices.index(keyslice) for keyslice in keyslices._slices])
             
            if(isinstance(values,tuple)):
                vslices = source.Get(*values)
            else:
                vslices = source.Get(values)
            
            pos = []
            for slice in vslices._slices:
                assert slice in oslices, "Cannot find slice specified in group flat argument"
                p = oslices.index(slice)
                pos.append(p)
                oslices[p] = None
            xflat[nkey] = set(pos)

        #group slices
        for pos, slice in enumerate(gsource._slices):
            p = oslices.index(slice)
            key = (pos,)
            if(key in xflat):
                xflat[key].add(p)
            else:
                xflat[key] = set([p])
            oslices[p] = None

        #filter remaining slices first
        pos = [p for p,s in enumerate(oslices) if not s is None]
        if(pos):
            fslices = [s for s in oslices if not s is None]
            fslices = Filter._apply(fslices,gslice,gslices[0].type.dims[:1],"dim")
            for p,fslice in zip(pos,fslices):
                nslices[p] = fslice

        #filter flattened slices 
        gidx = range(len(gsource._slices))
        gdims = gslice.dims[-len(gsource._slices):]
        firstelem = wrapper_py.Rep(0)._slices[0]
        for key, pos in xflat.iteritems():
            tslice = gslice
            for elem in gidx:
                if not elem in key:
                    selpath = dimpaths.DimPath(gdims[elem])
                    tslice = repops_funcs.Sum._apply([tslice],selpath)[0]
            tslice = ops.UnpackArrayOp(tslice,1)
            tslice = Filter._apply([tslice], firstelem,tslice.dims,"dim")[0]
            tslice = ops.PackArrayOp(tslice,1)
            pos = list(pos)
            fslices = [source._slices[p] for p in pos]
            fslices = Filter._apply(fslices,tslice,gslices[0].type.dims[:1],"dim")
            for p,fslice in zip(pos,fslices):
                nslices[p] = fslice
            
        return self._initialize(tuple(nslices),source._state)


class Filter(repops.MultiOpRep):
    def __init__(self,source,constraint,dim=None):
        if(not isinstance(constraint,representor.Representor)):
            constraint = repops.PlusPrefix(wrapper_py.Rep(constraint,name="filter"))
        repops.MultiOpRep.__init__(self,(source,constraint),dim=dim)

    def _process(self,sources,dim):
        source,constraint = sources
        if not source._state & RS_SLICES_KNOWN:
            return
        if not constraint._state & RS_TYPES_KNOWN:
            return
        assert len(constraint._slices) == 1, "Filter constraint should have 1 slice"
        cslice = constraint._slices[0]
        seldimpath = dimpaths.identifyUniqueDimPathSource(source, dim)
        if(not seldimpath and not isinstance(cslice.type, rtypes.TypeBool)):
            raise RuntimeError, "Attempting to perform filter on non-existing dimension"
        
        if(isinstance(constraint, repops.PlusPrefix)):
            mode = "pos"
        else:
            mode = "dim"

        nslices = self._apply(source._slices,cslice,seldimpath,mode)

        return self._initialize(tuple(nslices),source._state)

    @classmethod
    def _apply(cls,fslices,cslice,seldimpath, bcmode):
        if(isinstance(cslice.type,rtypes.TypeBool)):
            assert cslice.dims, "Constraint should have at least one dimension"
            ndim = dimensions.Dim(UNDEFINED,(True,) * (len(cslice.dims) -1),  False, name = "f" + cslice.dims[-1].name)
            dim_suffix = None
            seldimpath = cslice.dims[-1:]
            dimset = dimpaths.createDimSet([slice.dims for slice in fslices])
            assert seldimpath[0] in dimset, "Cannot find last dimension of boolean filter in filter source (" + str(cslice.dims) + ")"
            cslice = ops.PackArrayOp(cslice,1)
        else:
            assert seldimpath, "Filter dimpath is empty"

            if(isinstance(cslice.type, rtypes.TypeInteger)):
                ndim = None
            elif(isinstance(cslice.type, rtypes.TypeArray)):
                assert len(cslice.type.dims) == 1, "Filter array should be 1-dimensional"
                assert isinstance(cslice.type.subtypes[0], rtypes.TypeInteger) and \
                            not isinstance(cslice.type.subtypes[0], rtypes.TypeBool), \
                            "Multi-dimensional arrays cannot be used as filter. Please unpack the arrays."
                ndim = cslice.type.dims[0]
            elif(isinstance(cslice.type, rtypes.TypeSlice)):
                ndim = dimensions.Dim(UNDEFINED, (True,) * len(cslice.dims), False,name = "f" + seldimpath.strip()[-1].name)
            else:
                raise RuntimeError, "Unknown constraint type in filter: " + str(cslice.type)
        
        nslices = []
        for slice in fslices:
            slice = ops.filter(slice, cslice, seldimpath, ndim, bcmode)
            nslices.append(slice)
        return nslices

class Sort(repops.MultiOpRep):
    def __init__(self,source,constraint=None,descend=False):
        if(constraint is None):
            repops.MultiOpRep.__init__(self,(source,),descend=descend)
        else:
            if(not isinstance(constraint,representor.Representor)):
                constraint = repops.PlusPrefix(wrapper_py.Rep(constraint,name="filter"))
            repops.MultiOpRep.__init__(self,(source,constraint),descend=descend)

    def _process(self, sources, descend):
        if not any([s._state & RS_SLICES_KNOWN for s in sources]):
            return

        if(len(sources) == 1): #no explicit constraint, use data itself
            source = sources[0]
            if len(source._slices) > 1:
                constraint = source.Tuple()
            else:
                constraint = source
        else:
            source,constraint = sources 

        if len(constraint._slices) > 1:
            constraint = constraint.Tuple()
        
        #fixme: make it slice-only (remove rep)
        constraint = repops_funcs.Argsort(constraint, descend=descend)
        cslice = ops.PackArrayOp(constraint._slices[0])
        nslices = Filter._apply(source._slices, cslice, cslice.type.dims[:1],"dim")
        return self._initialize(tuple(nslices),source._state & constraint._state)


class Unique(repops.MultiOpRep):
    def __init__(self,source,constraint=None,descend=False):
        if(constraint is None):
            repops.MultiOpRep.__init__(self,(source,),descend=descend)
        else:
            if(not isinstance(constraint,representor.Representor)):
                constraint = repops.PlusPrefix(wrapper_py.Rep(constraint,name="filter"))
            repops.MultiOpRep.__init__(self,(source,constraint),descend=descend)

    def _process(self, sources, descend):
        if not any([s._state & RS_SLICES_KNOWN for s in sources]):
            return

        if(len(sources) == 1): #no explicit constraint, use data itself
            source = sources[0]
            if len(source._slices) > 1:
                constraint = source.Tuple()
            else:
                constraint = source
        else:
            source,constraint = sources 

        if len(constraint._slices) > 1:
            constraint = constraint.Tuple()
        
        #fixme: make it slice-only (remove rep)
        dpath = dimpaths.identifyUniqueDimPathSource(constraint,None)
        uconstraint = repops_funcs.Argunique(constraint)
        cslice = uconstraint._slices[0]
        nslices = Filter._apply(source._slices, cslice, dpath[-1:],"dim")
        return self._initialize(tuple(nslices),source._state & constraint._state)


class Join(repops.MultiOpRep):
    def __init__(self, lsource, rsource, constraint):
        if(isinstance(constraint,context.Context)):
            c = Combine(lsource,rsource)
            constraint = context._apply(constraint,c)
        repops.MultiOpRep.__init__(self,(lsource,rsource, constraint))

    def _process(self, sources):
        lsource, rsource, constraint = sources
        if not lsource._state & RS_SLICES_KNOWN or not rsource._state & RS_SLICES_KNOWN \
           or not constraint._state & RS_TYPES_KNOWN:
            return
        
        ldims = dimpaths.createDimSet([s.dims for s in lsource._slices])
        rdims = dimpaths.createDimSet([s.dims for s in rsource._slices])
        
        assert len(constraint._slices) == 1, "Constraint should have only one slice"
        cslice = constraint._slices[0]

        if not cslice.dims[-1] in rdims:
            rdims,ldims = ldims,rdims
            lsource,rsource = rsource,lsource
            reverse = True
            assert cslice.dims[-1] in rdims, "Last dimension of constraint should be in one of the sources"
        else:
            reverse = False
        rdimpath = dimpaths.DimPath(cslice.dims[-1])
        for dim in cslice.dims[::-1]:
            if dim in ldims:
                ldimpath = dimpaths.DimPath(dim)
                break
        else:
            raise RuntimeError, "No dimension found for one of the sources in constraint"

        leftpos = repops_funcs.Pos._apply([cslice], ldimpath)[0]
        rightpos = repops_funcs.Pos._apply([cslice], rdimpath)[0]
        
        filters = Filter._apply([leftpos,rightpos],cslice, None,"dim")
        leftflat,rightflat = repops_dim.Flat._apply(filters, filters[0].dims[-1:])

        leftflat = ops.PackArrayOp(leftflat)
        rightflat = ops.PackArrayOp(rightflat)

        lslices = Filter._apply(lsource._slices, leftflat, ldimpath, "dim")
        rslices = Filter._apply(rsource._slices, rightflat, rdimpath, "dim")

        if(reverse):
            nslices = Combine._apply(rslices,lslices)
        else:
            nslices = Combine._apply(lslices,rslices)
        return self._initialize(tuple(nslices),lsource._state & rsource._state)


class Match(repops.MultiOpRep):
    def __init__(self, lsource, rsource, lslice=None, rslice=None, jointype="inner"):
        assert jointype in set(["inner","left","right","full"]), "Jointype should be inner, left, right or full"
        if(isinstance(lslice,context.Context)):
            lslice = context._apply(lslice,lsource)
        if(isinstance(rslice,context.Context)):
            rslice = context._apply(rslice,rsource)
        repops.MultiOpRep.__init__(self,(lsource,rsource,lslice,rslice),jointype=jointype)

    def _process(self, sources, jointype):
        lsource, rsource, lslice,rslice = sources
        if not lsource._state & RS_SLICES_KNOWN or not rsource._state & RS_SLICES_KNOWN:
            return
         
        if(lslice is None):
            if(rslice is None):
                common_names = set(lsource.Names) & set(rsource.Names)
                if(len(common_names) != 1):
                    raise RuntimeError, "Cannot find a unique common named slice"
                name = common_names.pop()
                lslice = getattr(lsource,name)
                rslice = getattr(rsource,name)
            else:
                lslice = getattr(lsource,rslice.Names[0])
        elif(rslice is None):
            rslice = getattr(rsource,lslice.Names[0])
        assert len(rslice._slices) == 1, "rslice parameter in match should have only one slice"
        assert len(lslice._slices) == 1, "lslice parameter in match should have only one slice"

        self._sources = (lsource, rsource, lslice, rslice)
        if not lslice._state & RS_TYPES_KNOWN or not rslice._state & RS_TYPES_KNOWN:
            return
        
        lslice = lslice._slices[0]
        rslice = rslice._slices[0]

        lindex,rindex = ops.EquiJoinIndexOp(lslice,rslice, jointype=jointype).results

        lslices = list(lsource._slices)
        rslices = list(rsource._slices)
        if(lslice.name == rslice.name and jointype=="inner" and lslice in lslices and rslice in rslices):
            rslices.pop(rslices.index(rslice))

        lslices = Filter._apply(lslices, lindex, lslice.dims, "dim")
        rslices = Filter._apply(rslices, rindex, rslice.dims, "dim")


        nslices = Combine._apply(lslices,rslices)
        return self._initialize(tuple(nslices),lsource._state & rsource._state)


class Stack(repops.MultiOpRep):
    def __init__(self, *sources, **kwargs):
        repops.MultiOpRep.__init__(self,sources, **kwargs)

    def _process(self, sources, dim=None):
        state = reduce(operator.__and__,[source._state for source in sources])
        if not state & RS_SLICES_KNOWN:
            return

        slicelens = set([len(source._slices) for source in sources])
        assert len(slicelens) == 1, "Sources of stack should have same number of slices"
        nslice = slicelens.pop()

        seldimpaths = [] 
        for source in sources:
            seldimpaths.append(dimpaths.identifyUniqueDimPathSource(source, dim))

        nslices = []
        dimdepth = []
        ndim = None
        slicelists = [source._slices for source in sources]
        for slicecol in zip(*slicelists):
            packdepths = []
            ncol = []
            for slice,dpath in zip(slicecol, seldimpaths):
                lastpos = slice.dims.matchDimPath(dpath)
                packdepths.append(len(slice.dims) - lastpos[-1])
                ncol.append(ops.PackArrayOp(slice,packdepths[-1]))
            
            res = ncol[0]
            for slice in ncol[1:]:
                res = repops_funcs.Add._apply((res,slice),"pos")
            dimdepth = min(packdepths)
            res = ops.UnpackArrayOp(res,dimdepth)
            
            if(len(nslices) == 0):
                ndim = res.dims[-dimdepth]
            else:
                res = ops.ChangeDimOp(res, len(res.dims) - dimdepth, ndim)
            nslices.append(res)

        return self._initialize(tuple(nslices),RS_CHECK)



        



#after receipt by Ferdinand Jamitzky on activestate.com: http://code.activestate.com/recipes/384122-infix-operators/

class Infix:
    def __init__(self, function, kwargs={}):
        self.function = function
        self.kwargs = kwargs

    def __ror__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(other, x), self.kwargs)

    def __or__(self, other):
        return self.function(other,**self.kwargs)

    def __call__(self, *params, **kwargs):
        if(params):
            kwargs.update(self.kwargs)
            return self.function(*params, **kwargs)
        else:
            kwargs.update(self.kwargs)
            return Infix(self.function,kwargs)
            

class RevInfix:
    def __init__(self, function, kwargs={}):
        self.function = function
        self.kwargs = kwargs

    def __ror__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(x, other), self.kwargs)

    def __or__(self, other):
        return self.function(other,**self.kwargs)

    def __call__(self, *params, **kwargs):
        if(params):
            kwargs.update(self.kwargs)
            left,right = params[:2]
            params = params[2:]
            return self.function(right,left, *params, **kwargs)
        else:
            kwargs.update(self.kwargs)
            return Infix(self.function,kwargs)





UNDEFINED = -1


RS_TYPES_KNOWN = 1
RS_SLICES_KNOWN = 2
RS_ALL_KNOWN = RS_TYPES_KNOWN | RS_SLICES_KNOWN
RS_INFERRED = 8
RS_CHECK = 16


BCCOPY=0
BCENSURE=1
BCEXIST=2
BCNEW=3


NOVAL="NOVALUE" #alternative None for functions which also accept NOne as valid input
                #but still want some params to be optional



_delay_import_(globals(),"ibidas","*")

class Pre(object):
    def register(self, func):
        setattr(self,func.__name__, func)
predefined_sources = Pre()


def yeast_feats():
    """Returns table of yeast genome features,
       from:  http://downloads.yeastgenome.org/chromosomal_feature/SGD_features.tab
    """

    rtype = """[feats:*]<(sgdid=bytes, feat_type=bytes, feat_qual=bytes, feat_name=bytes, gene_name=bytes, 
                          gene_aliases=bytes, feat_parent_name=bytes, sgdid_alias=bytes, chromosome=bytes, 
                          start=bytes, stop=bytes, strand=bytes[1], genetic_pos=bytes, coordinate_version=bytes[10], 
                          sequence_version=bytes, description=bytes)"""
    
    res = Read(Fetch("http://downloads.yeastgenome.org/chromosomal_feature/SGD_features.tab"),dtype=rtype)

    splitfunc = lambda x: x.split("|")
    res = res.To(_.gene_aliases,  Do=_.Each(splitfunc, dtype="[aliases:~]<bytes").Elem()[_ != ""])
    res = res.To(_.start, _.stop, Do=_.Cast("int$"))
    res = res.To(_.genetic_pos,   Do=_.Cast("real64$"))
    return res.Copy()
predefined_sources.register(yeast_feats)  


def yeast_aliases(feats):
    """Given a yeast feats table (see ``yeast_feats``), returns a 
       two column table with the official feature names and a column with aliases.

       Can be used to map names to official names (with match)"""

    res = Stack(feats.Get(_.feat_name, _.gene_aliases).Flat(), 
                feats.Get(_.feat_name, _.gene_name), 
                feats.Get(_.feat_name, _.feat_name))
    res = res/("feat_name", "alias")
    res = res%"feat_aliases"
    res = res[_.alias != ""]
    return res.Unique().Copy()
predefined_sources.register(yeast_aliases)  

def in_memory_db():
    """Returns an empty in memory database"""
    return Connect("sqlite:///:memory:");
predefined_sources.register(in_memory_db)


def yeastract(url="http://www.yeastract.com/download/RegulationTwoColumnTable_Documented_20101213.tsv.gz"):
    """Downloads documented transcription factor regulation interactions from yeastract"""
    rtype = "[tftargets:*]<(trans_factor=bytes, target=bytes)"
    res = Read(Fetch(url),dtype=rtype)
    return res.Copy()
predefined_sources.register(yeastract)


def string_interactions(dburl, species="Saccharomyces cerevisiae"):
    """Given a Postgres db with String data, specified in dburl, and a species, returns all interactions and their score.

    The database data can be obtained from String.

    example url: "postgres://username:password@hostname:port/string_dbname"

    Use ``connect`` to access the whole database::
       
       #Get available species names: 
       >>> connect(dburl).items.species.offical_name

    """
    z = Connect(dburl)
    pyeast = z.items.proteins 
    inter  = z.items.species.Match(z.network.protein_protein_links)
    inter  = inter[_.official_name == species]
    inter  = inter.Match(pyeast//"left", _.protein_id_a, _.protein_id)
    inter  = inter.Match(pyeast//"right",_.protein_id_b, _.protein_id)
    return   inter.Get(_.left.preferred_name/"left", 
                       _.right.preferred_name/"right", 
                       _.combined_score) % "interactions"
predefined_sources.register(string_interactions)


def omim_genemap():
    """The omim genemap data"""
    rtype = """[omim:*]<(chr_map_entry_nr=bytes, month=bytes, day=bytes, year=bytes, location=bytes, gene_names=bytes, 
                       gene_status=bytes[1], title=bytes, f8=bytes, mim=bytes, method=bytes, comments=bytes, f12=bytes, 
                       disease1=bytes, disease2=bytes, disease3=bytes, mouse=bytes, reference=bytes)"""
    
    res = Read(Fetch("ftp://ftp.ncbi.nih.gov/repository/OMIM/genemap"),dtype=rtype)
    
    res = res.Get(HArray(_.disease1, _.disease2, _.disease3)[_ != " "]/"disease", "~")
    splitfunc = lambda x: x.split(", ")
    res = res.To(_.gene_names,  Do=_.Each(splitfunc, dtype="[symbols:~]<string").Elem()[_ != ""])
    res = res.To(_.method,      Do=_.Each(splitfunc, dtype="[methods:~]<string").Elem()[_ != ""])
    res = res.To(_.disease,     Do=_.Sum().Each(omim_disease_parse,"[diseases:~]<string").Elem()[_ != ""])
    res = res.To(_.day, _.month, _.year, Do=_.Cast("int$"))
    res = res.To(_.mim, Do=_.Cast("int"))

    return res.Without(_.f8, _.f12, _.disease1, _.disease2, _.disease3).Copy()
predefined_sources.register(omim_genemap)

def omim_disease_parse(x):
    x = x.replace('{','').replace('}','')
    elems =  x.split('; ')
    return [elem.split(', ')[0] for elem in elems]


def go_annotations(dburl, genus="Saccharomyces", species="cerevisiae"):
    """Accesses GO annotations in a MySQL database.

       Database data can be obtained from the geneontology website.

       example url: "mysql://username:password@hostname:port/go
    """
    go = Connect(dburl)
    g = go.species
    g = g.Match(go.gene_product,     _.id,                   _.species_id)
    g = g.Match(go.association,      _.gene_product.id,      _.gene_product_id)
    g = g.Match(go.graph_path,       _.association.term_id,  _.term2_id)
    g = g.Match(go.term//"annot",_.term1_id,             _.id)
    g = g.Match(go.term//"rel",  _.relationship_type_id, _.id)
    g = g[(_.genus==genus) & (_.species == species)]
    return g.Get(_.symbol, _.annot.name/"annotation", _.rel.name/"relation_type", _.distance)%"annotations"
predefined_sources.register(go_annotations)

    



"""
The ibidas module contains all main functions for working with ibidas objects.
"""

__all__ = ["Rep","Read","Connect","_",
           "Array","Tuple","Combine","HArray",
           "Stack",
           "Pos","Argsort",
           "Any","All",
           "Max","Min",
           "Argmin","Argmax",
           "Mean","Median",
           "Sum",
           "Count",
           "Broadcast","CreateType",
           "newdim","Missing",
           "Corr","In","Contains",
           "Fetch","Serve","Get",
           "Load","Save",'newdim',
           "Invert","Abs", "Negative",
           ]

from utils import delay_import
from utils.util import save_rep as Save, load_rep as Load
from utils.context import _
from utils.missing import Missing
from utils.infix import Infix,RevInfix
from itypes import createType as CreateType
from wrappers.wrapper_py import Rep
from wrappers.wrapper_tsv import TSVRepresentor
from wrappers.wrapper_sql import open_db
from representor import newdim
import repops_dim 
from repops_multi import Broadcast, Combine, Sort, Stack
import repops_slice 
import repops_funcs
from download_cache import DownloadCache
from pre import predefined_sources as Get
from server import Serve
from constants import *

Fetch = DownloadCache()
In = Infix(repops_funcs.Within)
Contains = Infix(repops_funcs.Contains)

Pos = repops.delayable(default_params="#")(repops_funcs.Pos)
Argsort = repops.delayable()(repops_funcs.Argsort)
Argmax = repops.delayable()(repops_funcs.Argmin)
Argmin = repops.delayable()(repops_funcs.Argmax)
Sum = repops.delayable()(repops_funcs.Sum)
Any = repops.delayable()(repops_funcs.Any)
All = repops.delayable()(repops_funcs.All)
Max = repops.delayable()(repops_funcs.Max)
Min = repops.delayable()(repops_funcs.Min)
Mean = repops.delayable()(repops_funcs.Mean)
Median = repops.delayable()(repops_funcs.Median)
Count = repops.delayable()(repops_funcs.Count)
Corr = repops.delayable()(repops_funcs.Corr)
Invert = repops.delayable()(repops_funcs.Invert)
Abs = repops.delayable()(repops_funcs.Abs)
Negative = repops.delayable()(repops_funcs.Negative)


HArray = repops.delayable(nsources=UNDEFINED)(repops_slice.HArray)
Tuple = repops.delayable(nsources=UNDEFINED)(repops_slice.Tuple)
Array = repops.delayable()(repops_dim.Array)

def Read(url, **kwargs):
    format = kwargs.pop('format','tsv')

    if(format == 'tsv'):
        return TSVRepresentor(url, **kwargs) 
    else:
        raise RuntimeError("Unknown format specified")

def Connect(url, **kwargs):
    format = kwargs.pop('format','db')
    if(format == "db"):
        return open_db(url, **kwargs)
    else:
        raise RuntimeError("Unknown format specified")

delay_import.perform_delayed_imports()
